from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 217,
                "output_tokens": 267,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 218,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=49 candidates_tokens_details=None prompt_token_count=217 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=217
)] thoughts_token_count=218 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=484 traffic_type=None\
""",
                "total_tokens": 484,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"rating": 7, "author": {"first_name": "Patrick", "last_name": "Rothfuss"}, "title": "THE NAME OF THE WIND"}'
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "rating": 7,
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                        "title": "THE NAME OF THE WIND",
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xaf\x06\x01r\xc8\xda|\xe6\xffnc\x18$`\xbd\n5o\xec\xf1\xff\xd4\xcep\x82\xd7Hf\xc4an\xbf\xb8\xa6\xc0\x1aa\x81\x1a\xb0\x98%\x07\\\xb0dC0\x81\xf20\xf4$\xda\x16X\xfc\xe3\x88}\x088fK&FmQ\xe5\xc8\x12\xf7{/\x1d\xaa\x94\xe2\xd2\x03\x1eDy\x10\x83YA4\xe6\xd4\x85h\x05wk\x1f\xa0\x9a\xce\rp\x98v\xf8\x87\xa7T"\xc5\x07\xdd\nM\xf5m-5Ft\x1e\x02;A\xd0\xa3q\x18\xcd\x7fh\xa9=\xa5|\\t\xcd.3y\xfc!\x88\xb0`X4\'A\xd3\x92\xac\x18\x0e\x9a@`*a\x1c\xa8\xee\x8fk\x1c\x8ad\xad\x8313\x94\xaf\x96\x1f\xc1,O #\x91\xa6\xf9\xd3\xd2r\xd3\x80\xec\xfdI.\x83\x00\x0b\x87\xf9w\xb9l\x9c\xa2\x84\xee\xef\x08\x9b\x84al\xe7\r\xab?\xcc\xee\x9bH\x87\x85\xcd<\x9eT\x16<\xf1\xdcp!f\xa4\x07m\xf75\xfb\xf9d\x8b\xa8n5\x80\xab\xe8|\xf0\xde/\xa0\x15\xf7\xc4\xa9\x0e*\xce\xf6\x17\x95\xb6\xab\r\x14\xc0H\xc6\xffF~a\x98\xe6l\'\x1ef3\xcc\xbdY\x8dF\x18|\x9d\xfb\xeb\x0c\xcb\x18\x86u 7\x9d\x879b\xc1\x88@\x17\xa3\x82\xdf\xc3)\x10\x9ftl\x01\xfcY\xd1\xad\xd6\xce\xd7U\xe9E\t\xb0\xc7\x80\xc8\x0c\x82O[B\x88\xda\x18\x19\xe0N\xe6Y\x8a\t[I\xa2\xca\x03F\xe0\xc6\xc6l\r\xe0\xc7\x899\xb6\xa8\x10Hj?^\x9d\n\x86\x07\xae\xab\xc6\\\xb4\xf0z\x98\x88\xe9\x14\xdfb}O|\xc8\xb3\xb8\x87P\xc1\x18k\x9bl\xda\x97\xa4\x1fmeY\xf3L\xdc\x82\xa1\x83\xafyfo\n\xd37\x9b\x0b\x19\xd2\x11\xd2&5G\t\x13\xdd-b\x88\x13Q\xac\xccL\x17\xf6\xc0Xt\x04~D\xa0\xb9C\xe0\xbfx+\xa41\xce\xb3\x0c\xd3\xd5B\xd5\x04\x9f\xe0\xca\xd0\x9a]\xb4>\xd4w\x8f\xfb\xd5\xaf\x7f\x157\xc9\xa2\x1e!\xe1>\xbc\xd6t\xaa\xc6$\xe4\xea\x16\xebZ1\xb9\xad\xf0n\x0et|\x9c%bt*\xf8\xe8\\5n\t\xef\xdc\\^T\xf5\x9b\xcf\xe8R[>y\xa0\xba\xd5C\xeb\x8b}\xc8+\xee\xb6\x96y\x94IsBHq\x8a\xc6V\x81\x92\x80~\x1d\xdd\xbf\xf2*\xaaT\xd2\x80P\xfc\xc02\x016LV\x95\xe6z\x1fc\x0cW\xe2\xe7\x9a"\x97\x19\x99\xde@\x9a\xa6\xf2\xd8\xfd\xf8\\\xb1\xe4\\Z\x05\xb3~\xa6]\xa4\xad\xd6\x13h3\xf0\xcb\xf4\xd1U9\x88>\x96W`[\xdbB\'\xed\x83\xe8\xa1\x97$\xbe%\xd4\x05\xcd!\x84g\x93\xe6\xbez\xcf\x8b\xf8DJ\xba\x03\xdb\x0c\xe2\xb2\xe1\x89\xd6\xd7%O\xa9\xa3\x16\xc5m\x89\x8b\xd5\xc1X\xa0\x8d\x07\xb7\xbd\xcd\xd4jK!\xf9\x7fP\xbaE|\x9a\x8c\x8c\xbc\xc3\xc1|\xe1\xe1\xe9\xa5\xd2\x02.\xe6\xf6\x7f\xaf\x0f\x04\xa9\xd6\xf8\xd7`\xde\xb2B\x1c\xd3\x19KX\x9f\xd7:\xbf\xd5F\x87\x9bo\x9af\xb5+p\xa3h\xa6#8\xd0\xd2\xcb\xe0\xc6\xed\x91=\xcf\x8b\xf5=\xd3\x1dNF\t5\xd8\xc2k\\\x90\xfa\xb1|zf\xd3gOm;u\x040.\x06\xdb\xe7\x8dp\xf5\x8bP\xf2\\\x8f\xbd\x9b\x97\x9a\x90YC\x87Y\x11_pL$\xc3\t\x04"\xeb|b\xea`0W#\xdd\x12-dD\t>\n"\x01r\xc8\xda|j$\x82\x99\x8b\x03\x145)\xc5\xe5\xe5\xe0IZ\xcb\xd5g\xf0\x07\xf3sY\xaf\x01\x96\xb5q\xeb',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 217,
                "output_tokens": 406,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 357,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=49 candidates_tokens_details=None prompt_token_count=217 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=217
)] thoughts_token_count=357 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=623 traffic_type=None\
""",
                "total_tokens": 623,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "THE NAME OF THE WIND", "rating": 7, "author": {"first_name": "Patrick", "last_name": "Rothfuss"}}'
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "THE NAME OF THE WIND",
                                        "rating": 7,
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xe9\n\x01r\xc8\xda|Eh\xa5\x9e\xb6\xa4`\x95\xb7\x9b\x90\xec\xb9o\xdbOo\xce\xf4\xb7^\x94\xcf*aB\xe0.\x17\x9f[@!n\xb2\x07\x12\xde`\x8a\x1c\xcc\xe9\x90]/\xad\xa6\x8c\x02w|\xc7Fz\xea\r)\xab\x0c^\xd4\x1d`\xb5;\xb4ZTe,\xa6\xd0\x0c\x04\xcbL\xf7\x1d$;:!c\xb8w~\xcb\x1b\xcc#IH\x81\x14x\x1e\xdb\x02E\x04\xa5\x87\xa1\xfd&\xc5\x96\xd4\x8d2"\xdb]+\x18\xb9O/\x94\x10\xab/<1A\x98O\x9e\xcaH\xed\xc2\x9e28trn\xaf\xde\xeb\xfc\xf1*\x1a\xbf>)q\xd4\x915f\x18\x15h\x1au\xf4\xc2>\xca_5\x92`\xb5z\xa5/\xe5\x9b\xf6\x19\xb4C\xa5\xec:\xdb$\x9a\xe2}\xd4\x03\x80\x04%\xaa\t\xf8\x1d4\x01\x91\x1f\xda\x98\x1e\x19%[1\xaf\x9e\xc0^\xdf(\xa6\x13\xc5\xdd\xd6AR\xad\x02\xed\xf1\x1dl\xc6C\xe9m,[\xda@\xc9\xda\xa8\xfdW?\xba\x9e\xd1\xa1@\xd1\x07\xa4\xc8uO\xa6\xca\x86\x99\r\xfa\xaa\xc52\xa06\x99\xe3\xd8|\x9c\x991\x077e\xc4|\x14\xbav\x08\xee\x04/\x8b\xca\x9aex;`<\x1b\xd1\xf0S/07\xbc\xec\xf9\xb1e\xe9\xb9\xf5"69M\xc4l&n\x8aX\x06\xa5K\xd0\xd9\x16\xeb\x83(}44\xd0\xc3\x8e\x82\xd6\x96\xd5u\xf5l\xcbj\xfa\x95\xda\xbe@\x04r}\xc2\xb6[\xeb\xa3k\xe4A\xc5\xbd\xa9H\xd2Kn\x8e\x1b\x8c\xbd\x946K\xc7\xff\xcf\xac\xa0\x89-,S\xbc\x9d\x13+\x15\xac\x12*4\xe7v\xc1\xb9\xe5\x9aV=p\xa6\xc2Z9t\xff\xbe8\x12\xc6\xee\xde\xdfi\xd8\xd5?\xd8R\xbfk\xd8\xd4sO\x85\xf1\x97\xf3`\x0f\x91\xed\xbf%\x1a\xa2\x81\xdd\xbe\xb5D\xd5\xe6m\x1e`\xa0Mh2)\x95\tqX\xf9\xbf\xd3\x90\xed\x9eW\x0e!<\x86\x9f\x19\xff\xb1w+cV\x8d\xc9LcM#\x1d\xfb\x19\x07\xbf\xb4\xcbZ|e\xa8tw\xc4\xdb\xf0\xa7\xb9i\x8e\xd0I\xfa\x07d\x06H\xcf\xea^\xd4hv\x18\x1a|PI\xbcf\xd9&\x08\x91D\x94\x1f\xc2\xcf\xb4+\x7f\xd0\xb8\x89#\xc0V\xbd-3\xa0X_\x95\xdeV2`*\xcd\xb2\x16S\x0b+\xec\x94`\x88\xf4\x1b\xc7\xf4\xde\x80\xde\xc2\xa6\xcf\xd8\x958Q\n\xe7#\xb9\xe0\x80\xb9D\x811\xc0\xeb\xdd\x91\x00l\x86G\xa1m\xfcB[\x1e\x7fE:\x06HO\x9d\xc7\x86\xedZ\xb1\xb9e\x94\x80Jm\xcc#\xc2Z\xbcY5\xa0\x02\':\xc1j\x07\x0e\xfe\x80\xcaX\xaa\xc5\xed\t\x12\xa9\xdd\xab\xf7\xed\xab\x00G_\n\xbf?o\x964\xb7\x01\xae\xcc-"\xbf y\xfb\x0b\xf3\xc6\x14\x143(\x11\x1b;\xf0\xe3T\xc1\x1b\xfc\x0c^i/\xa9\x1c\xda\xfb\x91\xea\xfdW<Gs[\xaa\x0e@\x8d3(v\xf6\xdflH\xd1\xa2\xcf\x0b\xdd\x10\x1f\xd7\xe9)U\xa5&}\xe6\xf6/\xdeB\xef`,\x8d\x93\xa5\xc3\xe2\x95\xe1\xf7\x17\x90\x16\xefC\xb1\xc7\x08\xeb\x9a"\x00\xd6+k\xe7\xf0\xe9\xce\x07\xac\xe1\xf1\xb8&r\x02\xc2\x16(\xbe\x951\xc0{\xf0\xa6\xd9\xb5I\xfd\x15\x80\x93\xc2\xe7e\t\xf8\xaa\xbd\xef0\x0e#nRZ\xcbs\xad(\xeb\xc6\xd1\x03}\x90\xaa\xf79\xe6p_\x7f\xe2\x00~?\xacRj\xd3\xa1\xfd\xe4\x10~\x1fr\xe1p\x87\xc8\x02`\xadb\x16\xe3\xeb\x15\xb4\xfa\x0bO\xbat\xd0\x88\x8c\x12\xe7/\xdey[Ob\x0b\x07\x95\xcc\x96\xabbj\xfd\xb4b2h\xb2&\xc3\xee\x17Y\t\xac\x0f\xc2\xd3\xc2\xb3\xdc\xf2#\xad\xe9\x8c\xcb\xc6{?\x8e\x03\xf5\xba\xd6\xdc\x90\x13S\x8f\xdf\xe8!\xfd\xca\x8a\xc5\x81Ad\xbc#\x83A\x95T\xd0\x84\xc0w$"M*bB B\xe5*p\x88\xc9\xcd\xba\x92\xe5S!N@\xc5\x16i\x0f.\x01\xb9\xc7\x15b\xcf\x0bG\x1d\xa6\xa5\xbb\x9e\xceP\x88\xb2\xcd\x8fl\x0f*\x91\xf94\x1d\xb5\x97r\x94\xfc\xb7\xd3\xbdW\xc8$\xac>\xb4\xa7\xe9\xbf\xb1\x98\xd4\x84U\xfe\x177]\xc9x\x08\x8b\xeekn\xa1\xdaJr\xa0\xa0dg\x1ay\xa9\xec\xea\xe8\xc1\xc4\xcf\x01\x85\xda\xc5>\xc7#\xec\xff\xc1\xd4\xdd\xef\xbf6\xd1\\\x97\xa5\x17\xbc\xdan\x84!P\x16w2\xff\xfe\xbf\xb1\xdeA7T\xd7\x8a\xae^f ZH\xf67+s\xb9\xae\x93Y\xf2d\xc6\xe72\xdc\r\x8a\xf5\x07#e9f7O\xff/\x15\xfc\xed\x06\xcd\xfc@\x91\xfcp\xc9\xf1\x98\xbd>mH\x9f\xe0\x1en\x10bG\x00\x1ay\xc8\xff\x1d\x11\xe7\n.@\x9f\tfw\xc8O\x0bT\x1f\x00\xb6\x13|^\xcd\xbc\x89r\x05\xe0W\xf0\xaeS\xef-\xc4\x15i\x10\xc9\x07&ZN\x04N\xdbK\n\xdf\xce\xfc,\xef\\\xb3\xe6>q{\xbb\xedV\xcc\x17\xa0I\xf3{S\x96N\xa4\xdd;\xd3\x1b\xe1\xd3\x13ic\xa5\xfc\x7f\x9b\t\xb2*\x8b\xbe\xe9C\x86\xe1\x02\x1c\xbeCyn\x88)2o\xdf\x89o\xa1\x81\xa3\xa7ZYm\xd0A\x82\x12\n\x0f\xb6F\xee\xe9\xff\x814\x8aB\xa1\xe8\xba\x11P(^H\x88x\xde\x9a]+\xd6\x19Z}^\xee1\xd7jbo\xecKI\x02\xc8\xd2`\x08\xe5\xf4e"k\xe4\xccW\xe2B]\x93\xe4\xab\x08*\xb8\xf4\x97\x14a\xc9\x88T\x9b]\xb0^\x1f\n^o\x82\x03\xfbz\xcfv\x86\x9b<>\xe5\xf7zY\xa7\xac\xbe_\xd68\x87\xa8\xfb*\xd1\xaf{\xa1\xa0\'Myb\x18\xc7\xa3`<\xe3c^fnL_"\xdd\x0fU\x07D\xe4%\xd6;\xa0\xc3\x93\xe9\x1a\xf1!;\x183O\xe7\x9c\xfc',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "THE NAME OF THE WIND", "author": {"first_name": "Patrick", "last_name": "Rothfuss"}, "rating": 7}'
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "THE NAME OF THE WIND",
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                        "rating": 7,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n$\x01r\xc8\xda|\xdb\xd0\xee\x17uk\xb6V\x92t\x91(8\x94%nH\xb7\x0c\xe1\xa5\x1e\x14n\x95G\xa6\x9dJ\x8b\xd3\ns\x01r\xc8\xda|\xbf\x0f\xac\x0b1o0\"=\x8c\xb7\x84\xcd\x17\xa5\xa6E\x9fmq\xc5\x04e\xcf\xde6d\xb0\xaa\x98\x97\x0c\xee\xbd\xf1+\x00\xa5\x7f[\xfc\xe1Q\xe0\xc3\xb3F\xb0\x7f\tT\xcd\xec\x1am\x15\xf6)\xd8t1REM\xfba\x99_\xe4s\x0fH\xc6Nx\xd3\x8d>\xa5r^\x83Lf\xe9-\xe4H\xa7B\xa3\x14\x94#F\xa4EoP\x18o\xdc\xaa\xf3\xc1B\x1d\xb2\x05~\n\xdc\x01\x01r\xc8\xda|\x82M\x16\xd9\xd1\x01\x16%\xadC;0r\x1f\xf0{\xe3\x8a\xda\xd8Mn+\xea\x17\xf4`L\xa5\xa3\x88\\\x17q\xb0\naB\x87\x13\xe5\xe8\xedV\x00\xedx\xda\xec.\xe6\x1b\"|\x11.\xb6Z\xb9\xc1c\x18\xc3\xa1\xbf%\x08ww\x05T\x00\x98x'UE\xdf\xca.R\xd2\x91\xccY\xa7\xb7\xcb\xc6^\x96\xd3v\x94/\xc8\xfdL\xa7\xbck\xd2\x0cxG\xeb=U\rZY\xb5%\xf9\xc8\xbbL\x97D\xc7\xa0^\xbc\x7fF[\x8aZ\x18 \x15\xbc\xb2\x86\x9b\x07\xb7\x18Uv\xd3\x98\xd5\xfc9\xcf\x8d\xf5\xd5\xd8W\xca@\xf8\xcf\x8d\xbft\xf7\x82\xfa\xc9A\x9c\x04\xcb\x95\xc7\x01K\x1c '\x11\xc9\xe2\x99\xdf\x9d\x1c\x88\x9a\xcb\xd8Y|\x8f\xf6db \r'4ZL\xb51\xe6k\xd7#\x9eL3\xe2\xb8\x12\xc6\xde,8M\x1e\xf7\n\xce\x01\x01r\xc8\xda|\x02\xfc,\x9d\xcb8\xf6\xf5\xabi\xfe|\x84\x11F\xb1p@\tBC\x0bBS\xb8E}\xf1\xdc\x8d-\xd2@\xca\xa4\xd1\x98\x13\nC3\x08\xb27+,\x13\xd2\x14N\x84cJ\xf2\xffz\xb3\xa7oX\xf1\x17\x80Y4!\xf9\x1c\xb6R\x8a\xc0.R^\x02a8(\xa7\x15\xa8\xdf\x98\x8b\x97\x19\xac\xea\x0b\n\x90\xe05\x8e\x11\xc9\x9fh\x07\xd1$,\xbdl4V\x18Y\xb1D\t|*\x16\xd9\xafG\x80\xf2\x87\x88*\xdd<{\x11\xdbV\xffB?Z\xad\xe9\x1f\x0f\xf3\xd5q\xa3\xb7\xf3\x9c\xfe\xea\x18\x8ay-!R?T\xcc\n\xd6\x8a\xe6\x0b\xf1\x8f\xacwAD\x00\xd7\xa7\x1f'{3\\\x85\x96\xc0\xe2\xae\xdd>\x17\x86\xc6\xeb\xcc\xfc\x86\x81\x99)Cl\xad=\xdbj/A\xf6\x98\n\xca\x01\x01r\xc8\xda|\xde\xfb\x96\xa7~\x05l~\xf0\xd6\xe9\xd8l9\x80\x9aQ\x89d?\xb3|tS\xbd\x9bS\x9fh\x13\x87\x9065\xedS\xf0\xf4\xd1g\x1bEbC7\x13\x97?\r\xabP\x95#Dmk\x98[]\x9bh\x896/2\x8d%\x03\xe7r\xc2q\xea\xfc}\xa8f\x89x\xab\x857|t\xe7\xd6\x8c$\xfd\xf7\x7f\xf5B\xab\x8e\x00@J\xe7\x1c\xf5Q\x0f\xb4\xb0Oo,\xc4\xf9\xce\x80-[\xf2\"\xa9O,6\xda\n\x99\xe7\xb7\xeeq\x1a\xda\x95+\xf5\n\xceM\x99T7E\x0c\xa4\xdb}\xe1H\xaf}\x88\x1a/\x81\xce&F\xf6\xfcom\xe7&}q\x9d*4\xd4}\x90\x8cw\tm\xa3\xaf\x1a\xaa\x84\x90\xdf\xd1\xab\xa7\x0bw\xba\xcfw\xdb\xfc\x87\x8bE\xd2_\xba8\x01\n\x9e\x01\x01r\xc8\xda|\xa5\xdb\xd4n\xc0\x87\xe8+\xa9\xb2K:\xaah{\xce^k]\xf4V2\x86\xaf\xa6\xc1>\x88\xb0p\xeb\xdev\xdf\xb2n(\xe4\xc5\x94\t\xd3\xbd[\xd9\x97\xb1\xb6u\x1f;y\xf9N+,\xdd\xc5\x9c\x1bCK\xc8v\x7f\xb9\x05\xd6:x\xdeIt\xc2.4}\xd5\x8am\xc6]\xaa}Vt\xc2\x08S\x0boT\x82\xe1\xb8\xd99O\xf7\xa9\x1d4\x04\xf2m\xdbX\xe8\xbc\xd8\xd5K\xce\xc3u\x9d\x13}\x03\x12\xce\xbd\xf9l\xd1-\xbc\xcb\x08\xe0\\\xddc\x87\xaas\x9c\xf5\x1biMm\xc5\x00\xccR!\xe0\xb6\x7f\xd1\xa8j",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "usage": {
                "input_tokens": 217,
                "output_tokens": 49,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 204,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 266,
            },
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "THE NAME OF THE WIND", "author": {"last_name": "Rothfuss", "first_name": "Patrick"}, "rating": 7}'
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "THE NAME OF THE WIND",
                                        "author": {
                                            "last_name": "Rothfuss",
                                            "first_name": "Patrick",
                                        },
                                        "rating": 7,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n$\x01r\xc8\xda|\x8dj\x92V{N\xe2\xb6p\xa7!9\xdc\x89\xdb\x13\xd9j\x07\xffE\xebj/\xf8d(\xe2a\xd4,\nr\x01r\xc8\xda|j\xc2\x9cE\xfe\xb6v\xbf\xd0sz\x11\xa4\xf6\x83!$\xae\xd6oz\xed\xd6\xe5\xd7\xd7.L0\x8b \xa5 \x91$\x147\xb9\x84!\x9c]\xb0\x1c@l\xc5\x197#\xaf\xec\x0b9\xc8\xe5\xc0\xdegE\xbe\xb5V?\t\xd7\x98\xeeLE110F\xed\xbf\xb1\x88-\xd6\xc5\x01\xfa\x86\xe0\x84Myx\x10\xbb\xd6\xe02\xf4\x00\xb9\xbe\xf4\x03Zo\xd1\x1d\x8c#\xd1M\xb1\n\xec\x01\x01r\xc8\xda|4Z\xc3\xb1\xa3s\x91\x12\xb4\xa0\x91S\x1f\xb8Y\xf2D\xb1M\x03y\x15L\x80 \xda<\x1e\x95x]\xec-\xa7<\xfd\xe8~\x00\x9e\xa0\x85^\xf5\x82\x8a@\xd8\xb3\xd2\xbe\xe5\xdb+\x90\x87l\xcbM\xff\xb9\xe1\xfb\x17\xc7\x8d\x82\x7fvnm&\xe8\x97\x84\x83&\xcd\xb2i\xd1\xde\x18\x95\xbb\x97\xeb\x8d~"/\x97\r\xa1Tw\xb6P\x07+Yjp\x86\xe2\xaf\xabH\xd0V\x94\xa6u\x9a\xa8\xf4\xa1\xe4\xcct9\x05\x80\xb7\xf1\x854\xca\xf5\xdaY+\xb3\x10ZA\xdfQl\xaf\xe2\xc0$\x1e?p;t\x0bl\tc\x861\xad\x0b?\xe8\xf1\xca\xacW\x84/\xa3=\xac\n\xe5\xba\x84\xd3}\x96\x18\r\x0f\x17\xde\x87\xd1\r}.\x92\x84n\x9cR[>i\x03z\xd7\xb3\xfd"\x1769\xd2\xb6\x88\xe5\xff\x84\x12\xe6[I\xdf\x04t\xec\x81\x96\xa6\xf3i\x1dd\x04\xd9\xf4\xaf\xc2\xe7\xe6\xe9\xae\n\xc0\x01\x01r\xc8\xda|\xc0\x11\x95\r{\x020o\x15\xc6\x7f\xeb\xec\xe30\xb0\x1b\x91\xc9\xe5\x9d3\xf3\xa4\xbb\xc8\x8d\xce\x15&\xaa\xbe\xe7\x87\x17\x0e\xcazra\x1f\x17\x86\x17\x8e\x9e\x87\xb1,c\x14\xc6*\xd5\xbe\xf06W\x00\xf2\x12\x92;\x95\xdd\n\xda:\xf9\r\x1e\xee\x83\xe0%\xa4`7\x98\x00\xf1\xf9\x1b\xd4\xf1\x92F\xd1\xb7~$L\x9bJ\x83x\x9dZ\xdc\rT\x01\x16\xa0Cf\xb3f\xb70e\x8e\xc3_\x89\xb3\x1a\x86\t\xfe\x95vb\x15\xd6\x1f\xa5\xccFW8*ZR1\xa2fz\x08\x187\x08\x88`z\x17\xb1j\xc7\x98\xe0\x9e\xe9\x87\xda\x10\rP~Z\x9a\xdd~\nN\x0e2\x1b\xd7\x9d\x15\xfe\r\x81\xad\xc9G\xbeo\xdc\'\xc5)\x12F\x11\x87\n\xf5\x01\x01r\xc8\xda|c\xfaDT\xcb\x92\xa5\xf28\xb6%\x06f=K\xd7\xa18\xc0e\x08\xa8\x13-Q\xd3\xae\xb1\xa6C\x17Bp\xf7>p}c\x81(\xca\x14h\x03\xae\x9d\xc7\xc1\xc4\x86\x8d,\x8b\x8a(\x13\x92\xa2u&\xd2\xc1\xd9\xbd[\x7f\xf8\xda7D\x8c\x10e\x90\xff\xf0|\x98\xb0\xdd\x86+\xe5\x89S\xb3\x1bt\t0\x00\xb5\xbc\xcd9FJ-\x18[\x00C\x86h\x16\xad\xabfHBWp\x8a1s\xc0\xa1O\xa8\x929\x0e\xc0\x0fQN\x0b\xdb\xaf\xe87\x9b\xfb\xc3s\xfd[9"53W"\xd5\x0fo\x0e\xe8\xa2\xbd\xa9\xd0N\xc3<v\x80Z\xff\x1a\x80#C\x83\xfc@\'\x85\x9fW\xc7d\x94\xa5\x02|\x89b\xae\t\xebSo\x82,\x8f1A\x0f)\x8e@\x17\x8f@(\x18\xc4\x17\xc4X\xf38\n\xe3\x01\x81y\n\x00\x9c\xd5\\\xe5\x95\x95\x1aC\xb5\xc4\xe1]\x852\x88\x17F\xc9\xfa\xb6\xec1\x9e\x18\xa9\x94\x88\xc6-\x9b\n\xb7\x01\x01r\xc8\xda|>oI\n=\xff\xc8U\xac\x1f\xa6R\xe4y\x8e\xe7\xd9\xe6\x1c\x94f\x8b\xbdz\x9f\xfa\xed\xc4gF\x14\x1b\xa6qwv\xee\xa8\x81\x92W\xab\xfa\x16\x17\x928\x86\x0b\xa7\xect\x8d+\xd2\xa0G\\\xba_\x8e\xd2\xc4 \xac\xbc\x8e\x82\xcfu-\xa5\x89\xe72G+\x0f\x19\xea\xffl\x08\x08\xce_\xf0\xf6\xa6\x8f\xea?i7\xfe\x08\x87\xe7\xda{\x8f\xa6\x8f\xe85B\x922\xc9\x83!?\xa1 ;p\'>\xe6\xb8\xaa\x082\xb5\xa9Z\xb9\xb8Z\xf4\x0e]\xbb\xb4\xaf\x1b\r\xfca\xd6D\xeb\x10\xf1\xee\xe6D\xe6:+5d[{\x00\xdam\xe8.\xa9\x98\x9a\x86\x13\xec\xfd\xbd<\xee\xe4\xbc\xe7\x08w*}S\xad\nJ\x01r\xc8\xda|\xe8j1\xb7D\xe3\xa2\x01\x98&a\x12\xdd\xd7C\x0c\xe3\xeb\x86Qu\x1f\xbfp\x8a/L\x18`\xad\x8e\xda\xafL\x81E\xdd\x80f\xde\x15\xa2\xf3\xd1\xfb#\x03\xd8\x17\x13\xc5\xca\x14\xae\xed\x97\xff\xe0\x08\xd3\x80"\xc1/\xc1\xf107\xf1',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "usage": {
                "input_tokens": 217,
                "output_tokens": 49,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 231,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 266,
            },
            "n_chunks": 3,
        }
    }
)



from mirascope import llm


def ask_question(question: str) -> llm.Response:
    model = llm.use_model("openai/gpt-4o")
    return model.call(question)


# Uses the default model (gpt-4o)
response = ask_question("What is 2 + 2?")

# Override with a different model at runtime
with llm.model("anthropic/claude-sonnet-4-5"):
    response = ask_question("What is 2 + 2?")  # Uses Claude instead



"""OpenTelemetry GenAI instrumentation for `mirascope.llm.Model` methods."""

from __future__ import annotations

import weakref
from collections.abc import AsyncIterator, Iterator, Sequence
from contextlib import AbstractContextManager
from functools import wraps
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
    overload,
)

from opentelemetry import trace as otel_trace

from .....llm import (
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncContextTools,
    AsyncResponse,
    AsyncStreamResponse,
    AsyncTools,
    Context,
    ContextResponse,
    ContextStreamResponse,
    ContextTools,
    DepsT,
    FormatSpec,
    FormattableT,
    Message,
    Model,
    Response,
    RootResponse,
    StreamResponse,
    StreamResponseChunk,
    Tools,
    UserContent,
)
from .....llm.messages import promote_to_messages, user
from .common import (
    FormatParam,
    SpanContext,
    attach_response,
    attach_response_async,
    record_dropped_params,
    start_model_span,
)

if TYPE_CHECKING:
    from opentelemetry.trace import Span


# =============================================================================
# Original method references and wrapped state flags
# =============================================================================

_ORIGINAL_MODEL_CALL = Model.call
_MODEL_CALL_WRAPPED = False
_ORIGINAL_MODEL_CALL_ASYNC = Model.call_async
_MODEL_CALL_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_CALL = Model.context_call
_MODEL_CONTEXT_CALL_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_CALL_ASYNC = Model.context_call_async
_MODEL_CONTEXT_CALL_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_STREAM = Model.stream
_MODEL_STREAM_WRAPPED = False
_ORIGINAL_MODEL_STREAM_ASYNC = Model.stream_async
_MODEL_STREAM_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_STREAM = Model.context_stream
_MODEL_CONTEXT_STREAM_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_STREAM_ASYNC = Model.context_stream_async
_MODEL_CONTEXT_STREAM_ASYNC_WRAPPED = False

# Resume method originals and flags
_ORIGINAL_MODEL_RESUME = Model.resume
_MODEL_RESUME_WRAPPED = False
_ORIGINAL_MODEL_RESUME_ASYNC = Model.resume_async
_MODEL_RESUME_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_RESUME = Model.context_resume
_MODEL_CONTEXT_RESUME_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_RESUME_ASYNC = Model.context_resume_async
_MODEL_CONTEXT_RESUME_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_RESUME_STREAM = Model.resume_stream
_MODEL_RESUME_STREAM_WRAPPED = False
_ORIGINAL_MODEL_RESUME_STREAM_ASYNC = Model.resume_stream_async
_MODEL_RESUME_STREAM_ASYNC_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_RESUME_STREAM = Model.context_resume_stream
_MODEL_CONTEXT_RESUME_STREAM_WRAPPED = False
_ORIGINAL_MODEL_CONTEXT_RESUME_STREAM_ASYNC = Model.context_resume_stream_async
_MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED = False


# =============================================================================
# Model.call instrumentation
# =============================================================================


@overload
def _instrumented_model_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: None = None,
) -> Response: ...


@overload
def _instrumented_model_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatSpec[FormattableT],
) -> Response[FormattableT]: ...


@overload
def _instrumented_model_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> Response | Response[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CALL)
def _instrumented_model_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatParam = None,
) -> Response | Response[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.call`."""
    messages = promote_to_messages(content)
    with start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
    ) as span_ctx:
        response = _ORIGINAL_MODEL_CALL(
            self,
            content,
            tools=tools,
            format=format,
        )
        if span_ctx.span is not None:
            attach_response(
                span_ctx.span,
                response,
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return response


def wrap_model_call() -> None:
    """Returns None. Replaces `Model.call` with the instrumented wrapper."""
    global _MODEL_CALL_WRAPPED
    if _MODEL_CALL_WRAPPED:
        return
    Model.call = _instrumented_model_call
    _MODEL_CALL_WRAPPED = True


def unwrap_model_call() -> None:
    """Returns None. Restores the original `Model.call` implementation."""
    global _MODEL_CALL_WRAPPED
    if not _MODEL_CALL_WRAPPED:
        return
    Model.call = _ORIGINAL_MODEL_CALL
    _MODEL_CALL_WRAPPED = False


# =============================================================================
# Model.call_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: None = None,
) -> AsyncResponse: ...


@overload
async def _instrumented_model_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatSpec[FormattableT],
) -> AsyncResponse[FormattableT]: ...


@overload
async def _instrumented_model_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> AsyncResponse | AsyncResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CALL_ASYNC)
async def _instrumented_model_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatParam = None,
) -> AsyncResponse | AsyncResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.call_async`."""
    messages = promote_to_messages(content)
    with start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=True,
    ) as span_ctx:
        response = await _ORIGINAL_MODEL_CALL_ASYNC(
            self,
            content,
            tools=tools,
            format=format,
        )
        if span_ctx.span is not None:
            await attach_response_async(
                span_ctx.span,
                response,
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return response


def wrap_model_call_async() -> None:
    """Returns None. Replaces `Model.call_async` with the instrumented wrapper."""
    global _MODEL_CALL_ASYNC_WRAPPED
    if _MODEL_CALL_ASYNC_WRAPPED:
        return
    Model.call_async = _instrumented_model_call_async
    _MODEL_CALL_ASYNC_WRAPPED = True


def unwrap_model_call_async() -> None:
    """Returns None. Restores the original `Model.call_async` implementation."""
    global _MODEL_CALL_ASYNC_WRAPPED
    if not _MODEL_CALL_ASYNC_WRAPPED:
        return
    Model.call_async = _ORIGINAL_MODEL_CALL_ASYNC
    _MODEL_CALL_ASYNC_WRAPPED = False


# =============================================================================
# Model.context_call instrumentation
# =============================================================================


@overload
def _instrumented_model_context_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: None = None,
) -> ContextResponse[DepsT, None]: ...


@overload
def _instrumented_model_context_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT],
) -> ContextResponse[DepsT, FormattableT]: ...


@overload
def _instrumented_model_context_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CONTEXT_CALL)
def _instrumented_model_context_call(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatParam = None,
) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_call`."""
    messages = promote_to_messages(content)
    with start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=True,
    ) as span_ctx:
        response = _ORIGINAL_MODEL_CONTEXT_CALL(
            self,
            content,
            ctx=ctx,
            tools=tools,
            format=format,
        )
        if span_ctx.span is not None:
            attach_response(
                span_ctx.span,
                response,
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return response


def wrap_model_context_call() -> None:
    """Returns None. Replaces `Model.context_call` with the instrumented wrapper."""
    global _MODEL_CONTEXT_CALL_WRAPPED
    if _MODEL_CONTEXT_CALL_WRAPPED:
        return
    Model.context_call = _instrumented_model_context_call
    _MODEL_CONTEXT_CALL_WRAPPED = True


def unwrap_model_context_call() -> None:
    """Returns None. Restores the original `Model.context_call` implementation."""
    global _MODEL_CONTEXT_CALL_WRAPPED
    if not _MODEL_CONTEXT_CALL_WRAPPED:
        return
    Model.context_call = _ORIGINAL_MODEL_CONTEXT_CALL
    _MODEL_CONTEXT_CALL_WRAPPED = False


# =============================================================================
# Model.context_call_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_context_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: None = None,
) -> AsyncContextResponse[DepsT, None]: ...


@overload
async def _instrumented_model_context_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT],
) -> AsyncContextResponse[DepsT, FormattableT]: ...


@overload
async def _instrumented_model_context_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CONTEXT_CALL_ASYNC)
async def _instrumented_model_context_call_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatParam = None,
) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_call_async`."""
    messages = promote_to_messages(content)
    with start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=True,
    ) as span_ctx:
        response = await _ORIGINAL_MODEL_CONTEXT_CALL_ASYNC(
            self,
            content,
            ctx=ctx,
            tools=tools,
            format=format,
        )
        if span_ctx.span is not None:
            await attach_response_async(
                span_ctx.span,
                response,
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return response


def wrap_model_context_call_async() -> None:
    """Returns None. Replaces `Model.context_call_async` with the instrumented wrapper."""
    global _MODEL_CONTEXT_CALL_ASYNC_WRAPPED
    if _MODEL_CONTEXT_CALL_ASYNC_WRAPPED:
        return
    Model.context_call_async = _instrumented_model_context_call_async
    _MODEL_CONTEXT_CALL_ASYNC_WRAPPED = True


def unwrap_model_context_call_async() -> None:
    """Returns None. Restores the original `Model.context_call_async` implementation."""
    global _MODEL_CONTEXT_CALL_ASYNC_WRAPPED
    if not _MODEL_CONTEXT_CALL_ASYNC_WRAPPED:
        return
    Model.context_call_async = _ORIGINAL_MODEL_CONTEXT_CALL_ASYNC
    _MODEL_CONTEXT_CALL_ASYNC_WRAPPED = False


# =============================================================================
# Stream span handler helpers
# =============================================================================


def _attach_stream_span_handlers(
    *,
    response: ContextStreamResponse[DepsT, FormattableT | None]
    | StreamResponse[FormattableT | None],
    span_cm: AbstractContextManager[SpanContext],
    span: Span,
    request_messages: Sequence[Message],
) -> None:
    """Returns None. Closes the span when streaming completes."""
    chunk_iterator: Iterator[StreamResponseChunk] = response._chunk_iterator

    response_ref = weakref.ref(response)
    closed = False

    def _close_span(
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        nonlocal closed
        if closed:
            return
        closed = True
        response_obj = response_ref()
        if response_obj is not None:
            attach_response(
                span,
                response_obj,
                request_messages=request_messages,
            )
        span_cm.__exit__(exc_type, exc, tb)

    def _wrapped_iterator() -> Iterator[StreamResponseChunk]:
        with otel_trace.use_span(span, end_on_exit=False):
            try:
                yield from chunk_iterator
            except Exception as exc:  # noqa: BLE001
                _close_span(type(exc), exc, exc.__traceback__)
                raise
            else:
                _close_span(None, None, None)
            finally:
                _close_span(None, None, None)

    response._chunk_iterator = _wrapped_iterator()


def _attach_async_stream_span_handlers(
    *,
    response: AsyncContextStreamResponse[DepsT, FormattableT | None],
    span_cm: AbstractContextManager[SpanContext],
    span: Span,
    request_messages: Sequence[Message],
) -> None:
    """Returns None. Closes the span when async streaming completes."""
    chunk_iterator: AsyncIterator[StreamResponseChunk] = response._chunk_iterator

    response_ref = weakref.ref(response)
    closed = False

    def _close_span(
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        nonlocal closed
        if closed:
            return
        closed = True
        response_obj = response_ref()
        if response_obj is not None:
            attach_response(
                span,
                response_obj,
                request_messages=request_messages,
            )
        span_cm.__exit__(exc_type, exc, tb)

    async def _wrapped_iterator() -> AsyncIterator[StreamResponseChunk]:
        try:
            async for chunk in chunk_iterator:
                yield chunk
        except Exception as exc:  # noqa: BLE001
            _close_span(type(exc), exc, exc.__traceback__)
            raise
        else:
            _close_span(None, None, None)
        finally:
            _close_span(None, None, None)

    response._chunk_iterator = _wrapped_iterator()


# =============================================================================
# Model.stream instrumentation
# =============================================================================


@overload
def _instrumented_model_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: None = None,
) -> StreamResponse: ...


@overload
def _instrumented_model_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatSpec[FormattableT],
) -> StreamResponse[FormattableT]: ...


@overload
def _instrumented_model_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> StreamResponse | StreamResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_STREAM)
def _instrumented_model_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: Tools | None = None,
    format: FormatParam = None,
) -> StreamResponse | StreamResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.stream`."""
    messages = promote_to_messages(content)
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        response = _ORIGINAL_MODEL_STREAM(
            self,
            content,
            tools=tools,
            format=format,
        )
        span_cm.__exit__(None, None, None)
        return response

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            response = _ORIGINAL_MODEL_STREAM(
                self,
                content,
                tools=tools,
                format=format,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_stream_span_handlers(
            response=response,
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return response


def wrap_model_stream() -> None:
    """Returns None. Replaces `Model.stream` with the instrumented wrapper."""
    global _MODEL_STREAM_WRAPPED
    if _MODEL_STREAM_WRAPPED:
        return
    Model.stream = _instrumented_model_stream
    _MODEL_STREAM_WRAPPED = True


def unwrap_model_stream() -> None:
    """Returns None. Restores the original `Model.stream` implementation."""
    global _MODEL_STREAM_WRAPPED
    if not _MODEL_STREAM_WRAPPED:
        return
    Model.stream = _ORIGINAL_MODEL_STREAM
    _MODEL_STREAM_WRAPPED = False


# =============================================================================
# Model.stream_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: None = None,
) -> AsyncStreamResponse: ...


@overload
async def _instrumented_model_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatSpec[FormattableT],
) -> AsyncStreamResponse[FormattableT]: ...


@overload
async def _instrumented_model_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_STREAM_ASYNC)
async def _instrumented_model_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    tools: AsyncTools | None = None,
    format: FormatParam = None,
) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.stream_async`."""
    messages = promote_to_messages(content)
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        response = await _ORIGINAL_MODEL_STREAM_ASYNC(
            self,
            content,
            tools=tools,
            format=format,
        )
        span_cm.__exit__(None, None, None)
        return response

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            response = await _ORIGINAL_MODEL_STREAM_ASYNC(
                self,
                content,
                tools=tools,
                format=format,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_async_stream_span_handlers(
            response=cast(
                "AsyncContextStreamResponse[Any, FormattableT | None]", response
            ),
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return response


def wrap_model_stream_async() -> None:
    """Returns None. Replaces `Model.stream_async` with the instrumented wrapper."""
    global _MODEL_STREAM_ASYNC_WRAPPED
    if _MODEL_STREAM_ASYNC_WRAPPED:
        return
    Model.stream_async = _instrumented_model_stream_async
    _MODEL_STREAM_ASYNC_WRAPPED = True


def unwrap_model_stream_async() -> None:
    """Returns None. Restores the original `Model.stream_async` implementation."""
    global _MODEL_STREAM_ASYNC_WRAPPED
    if not _MODEL_STREAM_ASYNC_WRAPPED:
        return
    Model.stream_async = _ORIGINAL_MODEL_STREAM_ASYNC
    _MODEL_STREAM_ASYNC_WRAPPED = False


# =============================================================================
# Model.context_stream instrumentation
# =============================================================================


@overload
def _instrumented_model_context_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: None = None,
) -> ContextStreamResponse[DepsT, None]: ...


@overload
def _instrumented_model_context_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT],
) -> ContextStreamResponse[DepsT, FormattableT]: ...


@overload
def _instrumented_model_context_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> (
    ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
): ...


@wraps(_ORIGINAL_MODEL_CONTEXT_STREAM)
def _instrumented_model_context_stream(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: ContextTools[DepsT] | None = None,
    format: FormatParam = None,
) -> ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_stream`."""
    messages = promote_to_messages(content)
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        response = _ORIGINAL_MODEL_CONTEXT_STREAM(
            self,
            content,
            ctx=ctx,
            tools=tools,
            format=format,
        )
        span_cm.__exit__(None, None, None)
        return response

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            response = _ORIGINAL_MODEL_CONTEXT_STREAM(
                self,
                content,
                ctx=ctx,
                tools=tools,
                format=format,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_stream_span_handlers(
            response=response,
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return response


def wrap_model_context_stream() -> None:
    """Returns None. Replaces `Model.context_stream` with the instrumented wrapper."""
    global _MODEL_CONTEXT_STREAM_WRAPPED
    if _MODEL_CONTEXT_STREAM_WRAPPED:
        return
    Model.context_stream = _instrumented_model_context_stream
    _MODEL_CONTEXT_STREAM_WRAPPED = True


def unwrap_model_context_stream() -> None:
    """Returns None. Restores the original `Model.context_stream` implementation."""
    global _MODEL_CONTEXT_STREAM_WRAPPED
    if not _MODEL_CONTEXT_STREAM_WRAPPED:
        return
    Model.context_stream = _ORIGINAL_MODEL_CONTEXT_STREAM
    _MODEL_CONTEXT_STREAM_WRAPPED = False


# =============================================================================
# Model.context_stream_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_context_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: None = None,
) -> AsyncContextStreamResponse[DepsT, None]: ...


@overload
async def _instrumented_model_context_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT],
) -> AsyncContextStreamResponse[DepsT, FormattableT]: ...


@overload
async def _instrumented_model_context_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> (
    AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT]
): ...


@wraps(_ORIGINAL_MODEL_CONTEXT_STREAM_ASYNC)
async def _instrumented_model_context_stream_async(
    self: Model,
    content: UserContent | Sequence[Message],
    *,
    ctx: Context[DepsT],
    tools: AsyncContextTools[DepsT] | None = None,
    format: FormatParam = None,
) -> (
    AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT]
):
    """Returns a GenAI-instrumented result of `Model.context_stream_async`."""
    messages = promote_to_messages(content)
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=tools,
        format=format,
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        response = await _ORIGINAL_MODEL_CONTEXT_STREAM_ASYNC(
            self,
            content,
            ctx=ctx,
            tools=tools,
            format=format,
        )
        span_cm.__exit__(None, None, None)
        return response

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            response = await _ORIGINAL_MODEL_CONTEXT_STREAM_ASYNC(
                self,
                content,
                ctx=ctx,
                tools=tools,
                format=format,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_async_stream_span_handlers(
            response=response,
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return response


def wrap_model_context_stream_async() -> None:
    """Returns None. Replaces `Model.context_stream_async` with the instrumented wrapper."""
    global _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED
    if _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED:
        return
    Model.context_stream_async = _instrumented_model_context_stream_async
    _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED = True


def unwrap_model_context_stream_async() -> None:
    """Returns None. Restores the original `Model.context_stream_async` implementation."""
    global _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED
    if not _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED:
        return
    Model.context_stream_async = _ORIGINAL_MODEL_CONTEXT_STREAM_ASYNC
    _MODEL_CONTEXT_STREAM_ASYNC_WRAPPED = False


# =============================================================================
# Model.resume instrumentation
# =============================================================================


@overload
def _instrumented_model_resume(
    self: Model,
    *,
    response: Response,
    content: UserContent,
) -> Response: ...


@overload
def _instrumented_model_resume(
    self: Model,
    *,
    response: Response[FormattableT],
    content: UserContent,
) -> Response[FormattableT]: ...


@overload
def _instrumented_model_resume(
    self: Model,
    *,
    response: Response | Response[FormattableT],
    content: UserContent,
) -> Response | Response[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_RESUME)
def _instrumented_model_resume(
    self: Model,
    *,
    response: Response | Response[FormattableT],
    content: UserContent,
) -> Response | Response[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.resume`."""
    messages = list(response.messages) + [user(content)]
    with start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
    ) as span_ctx:
        result = _ORIGINAL_MODEL_RESUME(
            self,
            response=response,
            content=content,
        )
        if span_ctx.span is not None:
            attach_response(
                span_ctx.span,
                cast("RootResponse[Any, FormattableT | None]", result),
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return result


def wrap_model_resume() -> None:
    """Returns None. Replaces `Model.resume` with the instrumented wrapper."""
    global _MODEL_RESUME_WRAPPED
    if _MODEL_RESUME_WRAPPED:
        return
    Model.resume = _instrumented_model_resume
    _MODEL_RESUME_WRAPPED = True


def unwrap_model_resume() -> None:
    """Returns None. Restores the original `Model.resume` implementation."""
    global _MODEL_RESUME_WRAPPED
    if not _MODEL_RESUME_WRAPPED:
        return
    Model.resume = _ORIGINAL_MODEL_RESUME
    _MODEL_RESUME_WRAPPED = False


# =============================================================================
# Model.resume_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_resume_async(
    self: Model,
    *,
    response: AsyncResponse,
    content: UserContent,
) -> AsyncResponse: ...


@overload
async def _instrumented_model_resume_async(
    self: Model,
    *,
    response: AsyncResponse[FormattableT],
    content: UserContent,
) -> AsyncResponse[FormattableT]: ...


@overload
async def _instrumented_model_resume_async(
    self: Model,
    *,
    response: AsyncResponse | AsyncResponse[FormattableT],
    content: UserContent,
) -> AsyncResponse | AsyncResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_RESUME_ASYNC)
async def _instrumented_model_resume_async(
    self: Model,
    *,
    response: AsyncResponse | AsyncResponse[FormattableT],
    content: UserContent,
) -> AsyncResponse | AsyncResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.resume_async`."""
    messages = list(response.messages) + [user(content)]
    with start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=True,
    ) as span_ctx:
        result = await _ORIGINAL_MODEL_RESUME_ASYNC(
            self,
            response=response,
            content=content,
        )
        if span_ctx.span is not None:
            await attach_response_async(
                span_ctx.span,
                cast("RootResponse[Any, FormattableT | None]", result),
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return result


def wrap_model_resume_async() -> None:
    """Returns None. Replaces `Model.resume_async` with the instrumented wrapper."""
    global _MODEL_RESUME_ASYNC_WRAPPED
    if _MODEL_RESUME_ASYNC_WRAPPED:
        return
    Model.resume_async = _instrumented_model_resume_async
    _MODEL_RESUME_ASYNC_WRAPPED = True


def unwrap_model_resume_async() -> None:
    """Returns None. Restores the original `Model.resume_async` implementation."""
    global _MODEL_RESUME_ASYNC_WRAPPED
    if not _MODEL_RESUME_ASYNC_WRAPPED:
        return
    Model.resume_async = _ORIGINAL_MODEL_RESUME_ASYNC
    _MODEL_RESUME_ASYNC_WRAPPED = False


# =============================================================================
# Model.context_resume instrumentation
# =============================================================================


@overload
def _instrumented_model_context_resume(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextResponse[DepsT, None],
    content: UserContent,
) -> ContextResponse[DepsT, None]: ...


@overload
def _instrumented_model_context_resume(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> ContextResponse[DepsT, FormattableT]: ...


@overload
def _instrumented_model_context_resume(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CONTEXT_RESUME)
def _instrumented_model_context_resume(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_resume`."""
    messages = list(response.messages) + [user(content)]
    with start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=True,
    ) as span_ctx:
        result = _ORIGINAL_MODEL_CONTEXT_RESUME(
            self,
            ctx=ctx,
            response=response,
            content=content,
        )
        if span_ctx.span is not None:
            attach_response(
                span_ctx.span,
                cast("RootResponse[Any, FormattableT | None]", result),
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return result


def wrap_model_context_resume() -> None:
    """Returns None. Replaces `Model.context_resume` with the instrumented wrapper."""
    global _MODEL_CONTEXT_RESUME_WRAPPED
    if _MODEL_CONTEXT_RESUME_WRAPPED:
        return
    Model.context_resume = _instrumented_model_context_resume
    _MODEL_CONTEXT_RESUME_WRAPPED = True


def unwrap_model_context_resume() -> None:
    """Returns None. Restores the original `Model.context_resume` implementation."""
    global _MODEL_CONTEXT_RESUME_WRAPPED
    if not _MODEL_CONTEXT_RESUME_WRAPPED:
        return
    Model.context_resume = _ORIGINAL_MODEL_CONTEXT_RESUME
    _MODEL_CONTEXT_RESUME_WRAPPED = False


# =============================================================================
# Model.context_resume_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_context_resume_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextResponse[DepsT, None],
    content: UserContent,
) -> AsyncContextResponse[DepsT, None]: ...


@overload
async def _instrumented_model_context_resume_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> AsyncContextResponse[DepsT, FormattableT]: ...


@overload
async def _instrumented_model_context_resume_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextResponse[DepsT, None]
    | AsyncContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]: ...


@wraps(_ORIGINAL_MODEL_CONTEXT_RESUME_ASYNC)
async def _instrumented_model_context_resume_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextResponse[DepsT, None]
    | AsyncContextResponse[DepsT, FormattableT],
    content: UserContent,
) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_resume_async`."""
    messages = list(response.messages) + [user(content)]
    with start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=True,
    ) as span_ctx:
        result = await _ORIGINAL_MODEL_CONTEXT_RESUME_ASYNC(
            self,
            ctx=ctx,
            response=response,
            content=content,
        )
        if span_ctx.span is not None:
            await attach_response_async(
                span_ctx.span,
                cast("RootResponse[Any, FormattableT | None]", result),
                request_messages=messages,
            )
            record_dropped_params(span_ctx.span, span_ctx.dropped_params)
        return result


def wrap_model_context_resume_async() -> None:
    """Returns None. Replaces `Model.context_resume_async` with the instrumented wrapper."""
    global _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED
    if _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED:
        return
    Model.context_resume_async = _instrumented_model_context_resume_async
    _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED = True


def unwrap_model_context_resume_async() -> None:
    """Returns None. Restores the original `Model.context_resume_async` implementation."""
    global _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED
    if not _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED:
        return
    Model.context_resume_async = _ORIGINAL_MODEL_CONTEXT_RESUME_ASYNC
    _MODEL_CONTEXT_RESUME_ASYNC_WRAPPED = False


# =============================================================================
# Model.resume_stream instrumentation
# =============================================================================


@overload
def _instrumented_model_resume_stream(
    self: Model,
    *,
    response: StreamResponse,
    content: UserContent,
) -> StreamResponse: ...


@overload
def _instrumented_model_resume_stream(
    self: Model,
    *,
    response: StreamResponse[FormattableT],
    content: UserContent,
) -> StreamResponse[FormattableT]: ...


@overload
def _instrumented_model_resume_stream(
    self: Model,
    *,
    response: StreamResponse | StreamResponse[FormattableT],
    content: UserContent,
) -> StreamResponse | StreamResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_RESUME_STREAM)
def _instrumented_model_resume_stream(
    self: Model,
    *,
    response: StreamResponse | StreamResponse[FormattableT],
    content: UserContent,
) -> StreamResponse | StreamResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.resume_stream`."""
    messages = list(response.messages) + [user(content)]
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        result = _ORIGINAL_MODEL_RESUME_STREAM(
            self,
            response=response,
            content=content,
        )
        span_cm.__exit__(None, None, None)
        return result

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            result = _ORIGINAL_MODEL_RESUME_STREAM(
                self,
                response=response,
                content=content,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_stream_span_handlers(
            response=cast(StreamResponse[FormattableT | None], result),
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return result


def wrap_model_resume_stream() -> None:
    """Returns None. Replaces `Model.resume_stream` with the instrumented wrapper."""
    global _MODEL_RESUME_STREAM_WRAPPED
    if _MODEL_RESUME_STREAM_WRAPPED:
        return
    Model.resume_stream = _instrumented_model_resume_stream
    _MODEL_RESUME_STREAM_WRAPPED = True


def unwrap_model_resume_stream() -> None:
    """Returns None. Restores the original `Model.resume_stream` implementation."""
    global _MODEL_RESUME_STREAM_WRAPPED
    if not _MODEL_RESUME_STREAM_WRAPPED:
        return
    Model.resume_stream = _ORIGINAL_MODEL_RESUME_STREAM
    _MODEL_RESUME_STREAM_WRAPPED = False


# =============================================================================
# Model.resume_stream_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_resume_stream_async(
    self: Model,
    *,
    response: AsyncStreamResponse,
    content: UserContent,
) -> AsyncStreamResponse: ...


@overload
async def _instrumented_model_resume_stream_async(
    self: Model,
    *,
    response: AsyncStreamResponse[FormattableT],
    content: UserContent,
) -> AsyncStreamResponse[FormattableT]: ...


@overload
async def _instrumented_model_resume_stream_async(
    self: Model,
    *,
    response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
    content: UserContent,
) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]: ...


@wraps(_ORIGINAL_MODEL_RESUME_STREAM_ASYNC)
async def _instrumented_model_resume_stream_async(
    self: Model,
    *,
    response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
    content: UserContent,
) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
    """Returns a GenAI-instrumented result of `Model.resume_stream_async`."""
    messages = list(response.messages) + [user(content)]
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        result = await _ORIGINAL_MODEL_RESUME_STREAM_ASYNC(
            self,
            response=response,
            content=content,
        )
        span_cm.__exit__(None, None, None)
        return result

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            result = await _ORIGINAL_MODEL_RESUME_STREAM_ASYNC(
                self,
                response=response,
                content=content,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_async_stream_span_handlers(
            response=cast(
                "AsyncContextStreamResponse[Any, FormattableT | None]", result
            ),
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return result


def wrap_model_resume_stream_async() -> None:
    """Returns None. Replaces `Model.resume_stream_async` with the instrumented wrapper."""
    global _MODEL_RESUME_STREAM_ASYNC_WRAPPED
    if _MODEL_RESUME_STREAM_ASYNC_WRAPPED:
        return
    Model.resume_stream_async = _instrumented_model_resume_stream_async
    _MODEL_RESUME_STREAM_ASYNC_WRAPPED = True


def unwrap_model_resume_stream_async() -> None:
    """Returns None. Restores the original `Model.resume_stream_async` implementation."""
    global _MODEL_RESUME_STREAM_ASYNC_WRAPPED
    if not _MODEL_RESUME_STREAM_ASYNC_WRAPPED:
        return
    Model.resume_stream_async = _ORIGINAL_MODEL_RESUME_STREAM_ASYNC
    _MODEL_RESUME_STREAM_ASYNC_WRAPPED = False


# =============================================================================
# Model.context_resume_stream instrumentation
# =============================================================================


@overload
def _instrumented_model_context_resume_stream(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextStreamResponse[DepsT, None],
    content: UserContent,
) -> ContextStreamResponse[DepsT, None]: ...


@overload
def _instrumented_model_context_resume_stream(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> ContextStreamResponse[DepsT, FormattableT]: ...


@overload
def _instrumented_model_context_resume_stream(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextStreamResponse[DepsT, None]
    | ContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> (
    ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
): ...


@wraps(_ORIGINAL_MODEL_CONTEXT_RESUME_STREAM)
def _instrumented_model_context_resume_stream(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: ContextStreamResponse[DepsT, None]
    | ContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]:
    """Returns a GenAI-instrumented result of `Model.context_resume_stream`."""
    messages = list(response.messages) + [user(content)]
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        result = _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM(
            self,
            ctx=ctx,
            response=response,
            content=content,
        )
        span_cm.__exit__(None, None, None)
        return result

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            result = _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM(
                self,
                ctx=ctx,
                response=response,
                content=content,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_stream_span_handlers(
            response=cast("ContextStreamResponse[Any, FormattableT | None]", result),
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return result


def wrap_model_context_resume_stream() -> None:
    """Returns None. Replaces `Model.context_resume_stream` with the instrumented wrapper."""
    global _MODEL_CONTEXT_RESUME_STREAM_WRAPPED
    if _MODEL_CONTEXT_RESUME_STREAM_WRAPPED:
        return
    Model.context_resume_stream = _instrumented_model_context_resume_stream
    _MODEL_CONTEXT_RESUME_STREAM_WRAPPED = True


def unwrap_model_context_resume_stream() -> None:
    """Returns None. Restores the original `Model.context_resume_stream` implementation."""
    global _MODEL_CONTEXT_RESUME_STREAM_WRAPPED
    if not _MODEL_CONTEXT_RESUME_STREAM_WRAPPED:
        return
    Model.context_resume_stream = _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM
    _MODEL_CONTEXT_RESUME_STREAM_WRAPPED = False


# =============================================================================
# Model.context_resume_stream_async instrumentation
# =============================================================================


@overload
async def _instrumented_model_context_resume_stream_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextStreamResponse[DepsT, None],
    content: UserContent,
) -> AsyncContextStreamResponse[DepsT, None]: ...


@overload
async def _instrumented_model_context_resume_stream_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> AsyncContextStreamResponse[DepsT, FormattableT]: ...


@overload
async def _instrumented_model_context_resume_stream_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> (
    AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT]
): ...


@wraps(_ORIGINAL_MODEL_CONTEXT_RESUME_STREAM_ASYNC)
async def _instrumented_model_context_resume_stream_async(
    self: Model,
    *,
    ctx: Context[DepsT],
    response: AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT],
    content: UserContent,
) -> (
    AsyncContextStreamResponse[DepsT, None]
    | AsyncContextStreamResponse[DepsT, FormattableT]
):
    """Returns a GenAI-instrumented result of `Model.context_resume_stream_async`."""
    messages = list(response.messages) + [user(content)]
    span_cm = start_model_span(
        self,
        messages=messages,
        tools=response.toolkit,
        format=cast(FormatParam, response.format),
        activate=False,
    )
    span_ctx = span_cm.__enter__()
    if span_ctx.span is None:
        result = await _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM_ASYNC(
            self,
            ctx=ctx,
            response=response,
            content=content,
        )
        span_cm.__exit__(None, None, None)
        return result

    try:
        with otel_trace.use_span(span_ctx.span, end_on_exit=False):
            result = await _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM_ASYNC(
                self,
                ctx=ctx,
                response=response,
                content=content,
            )
    except Exception as exc:
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    record_dropped_params(span_ctx.span, span_ctx.dropped_params)

    try:
        _attach_async_stream_span_handlers(
            response=cast(
                "AsyncContextStreamResponse[Any, FormattableT | None]", result
            ),
            span_cm=span_cm,
            span=span_ctx.span,
            request_messages=messages,
        )
    except Exception as exc:  # pragma: no cover
        span_cm.__exit__(type(exc), exc, exc.__traceback__)
        raise

    return result


def wrap_model_context_resume_stream_async() -> None:
    """Returns None. Replaces `Model.context_resume_stream_async` with the instrumented wrapper."""
    global _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED
    if _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED:
        return
    Model.context_resume_stream_async = _instrumented_model_context_resume_stream_async
    _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED = True


def unwrap_model_context_resume_stream_async() -> None:
    """Returns None. Restores the original `Model.context_resume_stream_async` implementation."""
    global _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED
    if not _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED:
        return
    Model.context_resume_stream_async = _ORIGINAL_MODEL_CONTEXT_RESUME_STREAM_ASYNC
    _MODEL_CONTEXT_RESUME_STREAM_ASYNC_WRAPPED = False



from mirascope import ops


# Client side: inject context into outgoing request headers
def make_request():
    headers: dict[str, str] = {}
    ops.inject_context(headers)
    # headers now contains trace context like:
    # {"traceparent": "00-...", "tracestate": "..."}
    print(f"Injected headers: {headers}")
    return headers


# Server side: extract context from incoming request headers
def handle_request(headers: dict[str, str]):
    with ops.propagated_context(extract_from=headers):
        # All traces within this block are linked to the parent trace
        process_request()


@ops.trace
def process_request() -> str:
    return "Request processed"


# Example usage
with ops.session(id="demo-session"):
    with ops.span("client_request"):
        headers = make_request()

    # Simulate server receiving the request
    handle_request(headers)



from mirascope import llm


@llm.prompt
def simple_prompt(topic: str):
    # Converts to a user message with one piece of text.
    return f"Tell me about {topic}."


@llm.prompt
def multimodal_prompt(image_url: str):
    # Converts to a user message with an image and a piece of text.
    return [
        llm.Image.from_url(image_url),
        "What's in this image?",
    ]


@llm.prompt
def chat_prompt(topic: str):
    # These messages will be used directly
    return [
        llm.messages.system("You are a helpful assistant."),
        llm.messages.user(f"Tell me about {topic}."),
    ]



"""Tests for primitive type utilities."""

from enum import Enum
from typing import Annotated, Literal, Optional

import pytest
from pydantic import BaseModel, Field, ValidationError

from mirascope.llm.formatting.primitives import create_wrapper_model, is_primitive_type


class Book(BaseModel):
    """A book model for testing."""

    title: str
    author: str


class Color(Enum):
    """An enum for testing."""

    RED = "red"
    BLUE = "blue"
    GREEN = "green"


def test_is_primitive_type_basic_primitives() -> None:
    """Test detection of basic primitive types."""
    assert is_primitive_type(str) is True
    assert is_primitive_type(int) is True
    assert is_primitive_type(float) is True
    assert is_primitive_type(bool) is True
    assert is_primitive_type(bytes) is True
    assert is_primitive_type(list) is True
    assert is_primitive_type(dict) is True
    assert is_primitive_type(set) is True
    assert is_primitive_type(tuple) is True


def test_is_primitive_type_generics() -> None:
    """Test detection of generic types."""
    assert is_primitive_type(list[int]) is True
    assert is_primitive_type(list[Book]) is True
    assert is_primitive_type(dict[str, int]) is True
    assert is_primitive_type(dict[str, Book]) is True
    assert is_primitive_type(set[str]) is True
    assert is_primitive_type(tuple[int, str]) is True


def test_is_primitive_type_enum() -> None:
    """Test detection of Enum types."""
    assert is_primitive_type(Color) is True


def test_is_primitive_type_literal() -> None:
    """Test detection of Literal types."""
    assert is_primitive_type(Literal["a", "b"]) is True
    assert is_primitive_type(Literal[1, 2, 3]) is True
    assert is_primitive_type(Literal["option1"]) is True


def test_is_primitive_type_union() -> None:
    """Test detection of Union types."""
    assert is_primitive_type(str | int) is True
    assert is_primitive_type(str | None) is True
    assert is_primitive_type(Optional[str]) is True  # noqa: UP045 # noqa: UP007 (expect error for testing)
    assert is_primitive_type(int | str | float) is True


def test_is_primitive_type_annotated() -> None:
    """Test detection of Annotated types."""
    assert is_primitive_type(Annotated[str, "description"]) is True
    assert is_primitive_type(Annotated[int, Field(ge=0)]) is True
    assert is_primitive_type(Annotated[list[Book], Field(description="Books")]) is True


def test_is_primitive_type_base_model() -> None:
    """Test that BaseModel subclasses are NOT primitives."""
    assert is_primitive_type(Book) is False
    assert is_primitive_type(BaseModel) is False


def test_is_primitive_type_none() -> None:
    """Test that None is NOT a primitive."""
    assert is_primitive_type(None) is False
    assert is_primitive_type(type(None)) is False


def test_create_wrapper_model_str() -> None:
    """Test wrapper creation for str."""
    wrapper = create_wrapper_model(str)

    assert wrapper.__name__ == "strOutput"

    assert hasattr(wrapper, "model_fields")
    assert "output" in wrapper.model_fields

    instance = wrapper(output="hello")
    assert instance.output == "hello"

    schema = wrapper.model_json_schema()
    assert "properties" in schema
    assert "output" in schema["properties"]


def test_create_wrapper_model_int() -> None:
    """Test wrapper creation for int."""
    wrapper = create_wrapper_model(int)

    assert wrapper.__name__ == "intOutput"

    instance = wrapper(output=42)
    assert instance.output == 42
    assert isinstance(instance.output, int)


def test_create_wrapper_model_list() -> None:
    """Test wrapper creation for bare list."""
    wrapper = create_wrapper_model(list)

    assert "list" in wrapper.__name__.lower()

    instance = wrapper(output=[1, 2, 3])
    assert instance.output == [1, 2, 3]


def test_create_wrapper_model_list_of_ints() -> None:
    """Test wrapper creation for list[int]."""
    wrapper = create_wrapper_model(list[int])

    instance = wrapper(output=[1, 2, 3])
    assert instance.output == [1, 2, 3]
    assert all(isinstance(x, int) for x in instance.output)

    with pytest.raises(ValidationError):
        wrapper(output=["not", "ints"])


def test_create_wrapper_model_list_of_models() -> None:
    """Test wrapper creation for list[Book]."""
    wrapper = create_wrapper_model(list[Book])

    books_data = [
        {"title": "Book 1", "author": "Author 1"},
        {"title": "Book 2", "author": "Author 2"},
    ]
    instance = wrapper(output=books_data)
    assert len(instance.output) == 2
    assert instance.output[0].title == "Book 1"
    assert isinstance(instance.output[0], Book)

    schema = wrapper.model_json_schema()
    assert "$defs" in schema or "definitions" in schema


def test_create_wrapper_model_dict() -> None:
    """Test wrapper creation for dict[str, int]."""
    wrapper = create_wrapper_model(dict[str, int])

    instance = wrapper(output={"a": 1, "b": 2})
    assert instance.output == {"a": 1, "b": 2}

    with pytest.raises(ValidationError):
        wrapper(output={"a": "not an int"})


def test_create_wrapper_model_dict_with_model_values() -> None:
    """Test wrapper creation for dict[str, Book]."""
    wrapper = create_wrapper_model(dict[str, Book])

    books_data = {
        "book1": {"title": "Title 1", "author": "Author 1"},
        "book2": {"title": "Title 2", "author": "Author 2"},
    }
    instance = wrapper(output=books_data)
    assert len(instance.output) == 2
    assert instance.output["book1"].title == "Title 1"
    assert isinstance(instance.output["book1"], Book)


def test_create_wrapper_model_annotated() -> None:
    """Test wrapper creation for Annotated types."""
    annotated_type = Annotated[str, Field(description="A description", min_length=1)]
    wrapper = create_wrapper_model(annotated_type)

    assert "str" in wrapper.__name__

    instance = wrapper(output="test")
    assert instance.output == "test"

    with pytest.raises(ValidationError):
        wrapper(output="")


def test_create_wrapper_model_set() -> None:
    """Test wrapper creation for set."""
    wrapper = create_wrapper_model(set[str])

    instance = wrapper(output={"a", "b", "c"})
    assert instance.output == {"a", "b", "c"}
    assert isinstance(instance.output, set)


def test_create_wrapper_model_tuple() -> None:
    """Test wrapper creation for tuple."""
    wrapper = create_wrapper_model(tuple[int, str])

    instance = wrapper(output=(1, "test"))
    assert instance.output == (1, "test")
    assert isinstance(instance.output, tuple)


def test_create_wrapper_model_union() -> None:
    """Test wrapper creation for Union types."""
    wrapper = create_wrapper_model(str | int)

    instance1 = wrapper(output="string")
    assert instance1.output == "string"

    instance2 = wrapper(output=42)
    assert instance2.output == 42


def test_create_wrapper_model_optional() -> None:
    """Test wrapper creation for Optional types."""
    wrapper = create_wrapper_model(str | None)

    instance1 = wrapper(output="value")
    assert instance1.output == "value"

    instance2 = wrapper(output=None)
    assert instance2.output is None


def test_create_wrapper_model_literal() -> None:
    """Test wrapper creation for Literal types."""
    wrapper = create_wrapper_model(Literal["a", "b", "c"])

    instance = wrapper(output="a")
    assert instance.output == "a"

    with pytest.raises(ValidationError):
        wrapper(output="d")


def test_create_wrapper_model_enum() -> None:
    """Test wrapper creation for Enum types."""
    wrapper = create_wrapper_model(Color)

    instance = wrapper(output="red")
    assert instance.output == Color.RED

    instance2 = wrapper(output=Color.BLUE)
    assert instance2.output == Color.BLUE


def test_create_wrapper_model_bytes() -> None:
    """Test wrapper creation for bytes."""
    wrapper = create_wrapper_model(bytes)

    instance = wrapper(output=b"hello")
    assert instance.output == b"hello"
    assert isinstance(instance.output, bytes)


def test_create_wrapper_model_nested_generics() -> None:
    """Test wrapper creation for nested generic types."""
    wrapper = create_wrapper_model(dict[str, list[int]])

    data = {"nums1": [1, 2, 3], "nums2": [4, 5, 6]}
    instance = wrapper(output=data)
    assert instance.output == data
    assert isinstance(instance.output["nums1"], list)


def test_create_wrapper_model_json_validation() -> None:
    """Test that wrapper model validates JSON correctly."""
    wrapper = create_wrapper_model(list[Book])

    json_str = '{"output": [{"title": "Book", "author": "Author"}]}'
    instance = wrapper.model_validate_json(json_str)
    assert len(instance.output) == 1
    assert instance.output[0].title == "Book"

    invalid_json = '[{"title": "Book", "author": "Author"}]'
    with pytest.raises(ValidationError):
        wrapper.model_validate_json(invalid_json)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .functions_list_by_env_response_functions_item import (
    FunctionsListByEnvResponseFunctionsItem,
)


class FunctionsListByEnvResponse(UniversalBaseModel):
    functions: typing.List[FunctionsListByEnvResponseFunctionsItem]
    total: float

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    ProviderToolUsage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="Jan 1, 2026 is a date in the past. Use the web search tool to lookup the price of bitcoin on that date."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
I was unable to find the specific Bitcoin price for January 1, 2026 in these search results. The results show current Bitcoin prices (around $87,000-$88,000 as of late January 2026) and discuss recent market movements, but don't provide the historical price data for New Year's Day.

Let me search more specifically for historical Bitcoin price data for that exact date.\
"""
                        ),
                        Text(
                            text="""\
Based on the search results, I found some information about Bitcoin's price around January 1, 2026:

"""
                        ),
                        Text(
                            text="The price briefly touched an intraday high of $94,352 after opening the year near $87,400 on Jan. 1"
                        ),
                        Text(
                            text="""\
, according to Bitcoin Magazine Pro data.

Additionally, another source indicates that \
"""
                        ),
                        Text(
                            text="The flagship coin dropped over 1% in the past 24 hours to trade at about $87,255 at press time"
                        ),
                        Text(text=" around New Year's Eve, and "),
                        Text(
                            text="The price of Bitcoin (BTC) has declined by 0.41% over the last 24 hours"
                        ),
                        Text(
                            text="""\
 on January 1, 2026.

So based on the available data, **Bitcoin opened 2026 near $87,400 on January 1, 2026**, and during the day it reached an intraday high of $94,352. The price showed some volatility around the New Year period, with reports of it trading around $87,255 on New Year's Eve and experiencing minor declines of around 0.41% to 1% in the 24-hour period around January 1st.\
"""
                        ),
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "type": "text",
                                "text": """\
I was unable to find the specific Bitcoin price for January 1, 2026 in these search results. The results show current Bitcoin prices (around $87,000-$88,000 as of late January 2026) and discuss recent market movements, but don't provide the historical price data for New Year's Day.

Let me search more specifically for historical Bitcoin price data for that exact date.\
""",
                            },
                            {
                                "type": "text",
                                "text": """\
Based on the search results, I found some information about Bitcoin's price around January 1, 2026:

""",
                            },
                            {
                                "type": "text",
                                "text": "The price briefly touched an intraday high of $94,352 after opening the year near $87,400 on Jan. 1",
                            },
                            {
                                "type": "text",
                                "text": """\
, according to Bitcoin Magazine Pro data.

Additionally, another source indicates that \
""",
                            },
                            {
                                "type": "text",
                                "text": "The flagship coin dropped over 1% in the past 24 hours to trade at about $87,255 at press time",
                            },
                            {"type": "text", "text": " around New Year's Eve, and "},
                            {
                                "type": "text",
                                "text": "The price of Bitcoin (BTC) has declined by 0.41% over the last 24 hours",
                            },
                            {
                                "type": "text",
                                "text": """\
 on January 1, 2026.

So based on the available data, **Bitcoin opened 2026 near $87,400 on January 1, 2026**, and during the day it reached an intraday high of $94,352. The price showed some volatility around the New Year period, with reports of it trading around $87,255 on New Year's Eve and experiencing minor declines of around 0.41% to 1% in the 24-hour period around January 1st.\
""",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please also look up the price of Ethereum on that date"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="Let me search more specifically for Ethereum's historical price data around January 1, 2026."
                        ),
                        Text(
                            text="""\
Based on the search results, I found specific information about Ethereum's price around January 1, 2026:

"""
                        ),
                        Text(
                            text="By December 31, 2025, ETH was near $2,970–$3,024, and on January 1–2, 2026, the price held above $3,000, showing a modest year-end rebound as markets opened 2026 on a balanced note."
                        ),
                        Text(
                            text="""\


Additionally, other sources provide more context about Ethereum's price movement around that period:

"""
                        ),
                        Text(
                            text="Ethereum opened around $3,354 on 1 January 2025, fell to close near $2,488 by 30 June, and then re-accelerated to end the year at about $2,968 on 31 December, before moving above $3,100 in early January 2026."
                        ),
                        Text(
                            text="""\


So to summarize both cryptocurrencies' prices on January 1, 2026:

- **Bitcoin**: Opened near $87,400 and reached an intraday high of $94,352
- **Ethereum**: Was in the range of $2,970–$3,024 at year-end 2025 and held above $3,000 on January 1-2, 2026, before moving above $3,100 in early January

Both cryptocurrencies showed some strength entering 2026, with Bitcoin experiencing more dramatic intraday volatility and Ethereum showing a steadier recovery above key psychological levels.\
"""
                        ),
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "type": "text",
                                "text": "Let me search more specifically for Ethereum's historical price data around January 1, 2026.",
                            },
                            {
                                "type": "text",
                                "text": """\
Based on the search results, I found specific information about Ethereum's price around January 1, 2026:

""",
                            },
                            {
                                "type": "text",
                                "text": "By December 31, 2025, ETH was near $2,970–$3,024, and on January 1–2, 2026, the price held above $3,000, showing a modest year-end rebound as markets opened 2026 on a balanced note.",
                            },
                            {
                                "type": "text",
                                "text": """\


Additionally, other sources provide more context about Ethereum's price movement around that period:

""",
                            },
                            {
                                "type": "text",
                                "text": "Ethereum opened around $3,354 on 1 January 2025, fell to close near $2,488 by 30 June, and then re-accelerated to end the year at about $2,968 on 31 December, before moving above $3,100 in early January 2026.",
                            },
                            {
                                "type": "text",
                                "text": """\


So to summarize both cryptocurrencies' prices on January 1, 2026:

- **Bitcoin**: Opened near $87,400 and reached an intraday high of $94,352
- **Ethereum**: Was in the range of $2,970–$3,024 at year-end 2025 and held above $3,000 on January 1-2, 2026, before moving above $3,100 in early January

Both cryptocurrencies showed some strength entering 2026, with Bitcoin experiencing more dramatic intraday volatility and Ethereum showing a steadier recovery above key psychological levels.\
""",
                            },
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [{"name": "web_search"}],
            "usage": {
                "input_tokens": 17,
                "output_tokens": 515,
                "cache_read_tokens": 16242,
                "cache_write_tokens": 23258,
                "reasoning_tokens": 0,
                "provider_tool_usage": [
                    ProviderToolUsage(name="web_search", call_count=2)
                ],
                "raw": "None",
                "total_tokens": 532,
            },
            "n_chunks": 148,
        }
    }
)



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "exception": {
            "type": "FeatureNotSupportedError",
            "args": "('Google does not support URL-referenced images. Try `llm.Image.download(...) or `llm.Image.download_async(...)`',)",
            "feature": "url_image_source",
            "model_id": "None",
            "provider_id": "google",
        }
    }
)



"""OpenTelemetry GenAI instrumentation for `mirascope.llm`.

This module provides the public instrument_llm() and uninstrument_llm() functions
that enable/disable instrumentation for Model calls and Response.resume methods.
"""

from __future__ import annotations

import os

from opentelemetry import trace as otel_trace

from ...configuration import get_tracer
from .model import (
    unwrap_model_call,
    unwrap_model_call_async,
    unwrap_model_context_call,
    unwrap_model_context_call_async,
    unwrap_model_context_resume,
    unwrap_model_context_resume_async,
    unwrap_model_context_resume_stream,
    unwrap_model_context_resume_stream_async,
    unwrap_model_context_stream,
    unwrap_model_context_stream_async,
    unwrap_model_resume,
    unwrap_model_resume_async,
    unwrap_model_resume_stream,
    unwrap_model_resume_stream_async,
    unwrap_model_stream,
    unwrap_model_stream_async,
    wrap_model_call,
    wrap_model_call_async,
    wrap_model_context_call,
    wrap_model_context_call_async,
    wrap_model_context_resume,
    wrap_model_context_resume_async,
    wrap_model_context_resume_stream,
    wrap_model_context_resume_stream_async,
    wrap_model_context_stream,
    wrap_model_context_stream_async,
    wrap_model_resume,
    wrap_model_resume_async,
    wrap_model_resume_stream,
    wrap_model_resume_stream_async,
    wrap_model_stream,
    wrap_model_stream_async,
)
from .response import (
    unwrap_async_context_response_resume,
    unwrap_async_context_stream_response_resume,
    unwrap_async_response_resume,
    unwrap_async_stream_response_resume,
    unwrap_context_response_resume,
    unwrap_context_stream_response_resume,
    unwrap_response_resume,
    unwrap_stream_response_resume,
    wrap_async_context_response_resume,
    wrap_async_context_stream_response_resume,
    wrap_async_response_resume,
    wrap_async_stream_response_resume,
    wrap_context_response_resume,
    wrap_context_stream_response_resume,
    wrap_response_resume,
    wrap_stream_response_resume,
)


def instrument_llm() -> None:
    """Enable GenAI 1.38 span emission for future `llm.Model` calls and streams.

    Uses the tracer provider configured via `ops.configure()`. If no provider
    was configured, uses the global OpenTelemetry tracer provider.

    Example:

        Enable instrumentation with a custom provider:
        ```python
        from mirascope import ops
        from opentelemetry.sdk.trace import TracerProvider

        provider = TracerProvider()
        ops.configure(tracer_provider=provider)
        ops.instrument_llm()
        ```
    """
    if otel_trace is None:  # pragma: no cover
        raise ImportError(
            "OpenTelemetry is not installed. Run `pip install mirascope[otel]` "
            "and ensure `opentelemetry-api` is available before calling "
            "`instrument_llm`."
        )

    os.environ.setdefault("OTEL_SEMCONV_STABILITY_OPT_IN", "gen_ai_latest_experimental")

    if get_tracer() is None:  # pragma: no cover
        raise RuntimeError(
            "You must call `configure()` before calling `instrument_llm()`."
        )

    # Model call methods
    wrap_model_call()
    wrap_model_call_async()
    wrap_model_context_call()
    wrap_model_context_call_async()
    wrap_model_stream()
    wrap_model_stream_async()
    wrap_model_context_stream()
    wrap_model_context_stream_async()

    # Model resume methods
    wrap_model_resume()
    wrap_model_resume_async()
    wrap_model_context_resume()
    wrap_model_context_resume_async()
    wrap_model_resume_stream()
    wrap_model_resume_stream_async()
    wrap_model_context_resume_stream()
    wrap_model_context_resume_stream_async()

    # Response resume methods (Mirascope-specific spans)
    wrap_response_resume()
    wrap_async_response_resume()
    wrap_context_response_resume()
    wrap_async_context_response_resume()
    wrap_stream_response_resume()
    wrap_async_stream_response_resume()
    wrap_context_stream_response_resume()
    wrap_async_context_stream_response_resume()


def uninstrument_llm() -> None:
    """Disable previously configured instrumentation."""
    # Model call methods
    unwrap_model_call()
    unwrap_model_call_async()
    unwrap_model_context_call()
    unwrap_model_context_call_async()
    unwrap_model_stream()
    unwrap_model_stream_async()
    unwrap_model_context_stream()
    unwrap_model_context_stream_async()

    # Model resume methods
    unwrap_model_resume()
    unwrap_model_resume_async()
    unwrap_model_context_resume()
    unwrap_model_context_resume_async()
    unwrap_model_resume_stream()
    unwrap_model_resume_stream_async()
    unwrap_model_context_resume_stream()
    unwrap_model_context_resume_stream_async()

    # Response resume methods (Mirascope-specific spans)
    unwrap_response_resume()
    unwrap_async_response_resume()
    unwrap_context_response_resume()
    unwrap_async_context_response_resume()
    unwrap_stream_response_resume()
    unwrap_async_stream_response_resume()
    unwrap_context_stream_response_resume()
    unwrap_async_context_stream_response_resume()



"""Google error handling utilities."""

from google.genai.errors import (
    ClientError as GoogleClientError,
    ServerError as GoogleServerError,
)

from ....exceptions import (
    APIError,
    AuthenticationError,
    BadRequestError,
    NotFoundError,
    PermissionError,
    ProviderError,
    RateLimitError,
    ServerError,
)
from ...base import ProviderErrorMap


def map_google_error(e: Exception) -> type[ProviderError]:
    """Map Google error to appropriate Mirascope error type.

    Google only provides ClientError (4xx) and ServerError (5xx) with status codes,
    so we map based on status code and message patterns.
    """
    if not isinstance(e, GoogleClientError | GoogleServerError):
        return ProviderError

    # Authentication errors (401) or 400 with "API key not valid"
    if e.code == 401 or (e.code == 400 and "API key not valid" in str(e)):
        return AuthenticationError
    if e.code == 403:
        return PermissionError
    if e.code == 404:
        return NotFoundError
    if e.code == 429:
        return RateLimitError
    if e.code in (400, 422):
        return BadRequestError
    if isinstance(e, GoogleServerError) and e.code >= 500:
        return ServerError
    return APIError


# Shared error mapping for Google provider
GOOGLE_ERROR_MAP: ProviderErrorMap = {
    GoogleClientError: map_google_error,
    GoogleServerError: map_google_error,
}



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .organization_memberships_list_response_item import (
    OrganizationMembershipsListResponseItem,
)
from .organization_memberships_list_response_item_role import (
    OrganizationMembershipsListResponseItemRole,
)
from .organization_memberships_update_request_role import (
    OrganizationMembershipsUpdateRequestRole,
)
from .organization_memberships_update_response import (
    OrganizationMembershipsUpdateResponse,
)
from .organization_memberships_update_response_role import (
    OrganizationMembershipsUpdateResponseRole,
)

__all__ = [
    "OrganizationMembershipsListResponseItem",
    "OrganizationMembershipsListResponseItemRole",
    "OrganizationMembershipsUpdateRequestRole",
    "OrganizationMembershipsUpdateResponse",
    "OrganizationMembershipsUpdateResponseRole",
]



from typing import Any, Protocol

from ..context import Context, DepsT
from ..types import JsonableCovariantT, P


class ToolFn(Protocol[P, JsonableCovariantT]):
    """Protocol for the tool function."""

    __name__: str
    """The name of the tool."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> JsonableCovariantT:
        """Call the function with the given arguments."""
        raise NotImplementedError()


class AsyncToolFn(Protocol[P, JsonableCovariantT]):
    """Protocol for the async tool function."""

    __name__: str
    """The name of the tool."""

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> JsonableCovariantT:
        """Call the function with the given arguments."""
        raise NotImplementedError()


class ContextToolFn(Protocol[DepsT, P, JsonableCovariantT]):
    """Protocol for the context tool function."""

    __name__: str
    """The name of the tool."""

    def __call__(
        self, ctx: Context[DepsT], *args: P.args, **kwargs: P.kwargs
    ) -> JsonableCovariantT:
        """Call the function with the given arguments."""
        raise NotImplementedError()


class AsyncContextToolFn(Protocol[DepsT, P, JsonableCovariantT]):
    """Protocol for the async context tool function."""

    __name__: str
    """The name of the tool."""

    async def __call__(
        self, ctx: Context[DepsT], *args: P.args, **kwargs: P.kwargs
    ) -> JsonableCovariantT:
        """Call the function with the given arguments."""
        raise NotImplementedError()


class KwargsCallable(Protocol[JsonableCovariantT]):
    """Protocol for functions that can be called with `Any`-typed kwargs.

    Used internally to type-cast tool functions for compatibility with
    json.loads() output when executing tool calls.
    """

    def __call__(self, **kwargs: dict[str, Any]) -> JsonableCovariantT: ...


class AsyncKwargsCallable(Protocol[JsonableCovariantT]):
    """Protocol for async functions that can be called with `Any`-typed kwargs.

    Used internally to type-cast async tool functions for compatibility with
    json.loads() output when executing tool calls.
    """

    async def __call__(self, **kwargs: dict[str, Any]) -> JsonableCovariantT: ...


class ContextKwargsCallable(Protocol[DepsT, JsonableCovariantT]):
    """Protocol for context functions that can be called with `Any`-typed kwargs.

    Used internally to type-cast context tool functions for compatibility with
    json.loads() output when executing tool calls.
    """

    def __call__(
        self, ctx: Context[DepsT], **kwargs: dict[str, Any]
    ) -> JsonableCovariantT: ...


class AsyncJsonKwargsCallable(Protocol[DepsT, JsonableCovariantT]):
    """Protocol for async context functions that can be called with `Any`-typed kwargs.

    Used internally to type-cast async context tool functions for compatibility with
    json.loads() output when executing tool calls.
    """

    async def __call__(
        self, ctx: Context[DepsT], **kwargs: dict[str, Any]
    ) -> JsonableCovariantT: ...



"""Tests for LLM instrumentation control (instrument/uninstrument, tracer context)."""

from __future__ import annotations

from collections.abc import Generator

import pytest
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)

from mirascope import llm, ops
from mirascope.ops._internal import configuration as ops_configuration
from mirascope.ops._internal.configuration import set_tracer


@pytest.fixture(autouse=True, scope="function")
def initialize(tracer_provider: TracerProvider) -> Generator[None, None, None]:
    """Initialize ops configuration and LLM instrumentation for each test."""
    ops.configure(tracer_provider=tracer_provider)
    ops.instrument_llm()
    yield
    ops.uninstrument_llm()


def test_uninstrument_llm() -> None:
    """Test uninstrumentation and re-instrumentation is idempotent."""
    ops.uninstrument_llm()

    ops.instrument_llm()
    first_call = llm.Model.call
    ops.instrument_llm()
    assert llm.Model.call is first_call

    ops.uninstrument_llm()


@pytest.mark.vcr()
def test_model_call_without_instrumentation(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that model.call works without instrumentation (no spans created)."""
    ops.uninstrument_llm()

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("Hello")]

    response = model.call(messages)

    assert response.content
    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0


@pytest.mark.vcr()
def test_model_call_with_tracer_set_to_none(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that span() returns None when tracer is explicitly set to None."""
    set_tracer(None)

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("Test with None tracer")]

    response = model.call(messages)

    assert response.content
    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0


def test_tracer_context_temporarily_sets_tracer(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that tracer_context temporarily sets a tracer and restores the previous."""
    original_tracer = ops_configuration.get_tracer()

    provider = TracerProvider()
    new_tracer = provider.get_tracer("test-tracer")

    with ops.tracer_context(new_tracer) as ctx_tracer:
        assert ctx_tracer is new_tracer
        assert ops_configuration.get_tracer() is new_tracer

    assert ops_configuration.get_tracer() is original_tracer


@pytest.mark.vcr()
def test_tracer_context_with_model_call(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that tracer_context works with model calls."""
    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("Say hello")]

    with ops.tracer_context(None):
        response = model.call(messages)
        assert response.content

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0



"""Mirascope OpenTelemetry exporters for two-phase telemetry.

This package provides a two-phase export system for OpenTelemetry tracing:
1. Immediate start event transmission for real-time visibility
2. Batched end event transmission for efficiency
"""

from .exporters import MirascopeOTLPExporter
from .processors import MirascopeSpanProcessor
from .types import (
    Link,
    SpanContextDict,
    SpanEvent,
    SpanEventType,
    Status,
)

__all__ = [
    "Link",
    "MirascopeOTLPExporter",
    "MirascopeSpanProcessor",
    "SpanContextDict",
    "SpanEvent",
    "SpanEventType",
    "Status",
]



"""End-to-end tests for LLM call with thinking=True parameter."""

import pytest

from mirascope import llm
from tests.utils import (
    Snapshot,
    snapshot_test,
)

E2E_MODEL_IDS: list[llm.ModelId] = [
    "openai/gpt-5:responses",
    "anthropic/claude-sonnet-4-0",
    "anthropic-beta/claude-sonnet-4-0",
    "google/gemini-2.5-flash",
]

# This can't be easily answered without thinking
PROMPT = "How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
# Check whether the model thinks it retained access to its reasoning process
RESUME_PROMPT = (
    "If you remember what the primes were, then share them, or say 'I don't remember.'"
)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_thinking_sync(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test synchronous call with thinking=True to verify reasoning content."""

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": True})
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call(PROMPT)
        with llm.model(model_id, thinking=llm.ThinkingConfig(level="minimal")):
            response = response.resume(RESUME_PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_thinking_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test streaming call with thinking=True to verify reasoning content."""

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": True})
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call.stream(PROMPT)
        response.finish()
        with llm.model(model_id, thinking=llm.ThinkingConfig(level="minimal")):
            response = response.resume(RESUME_PROMPT)
        response.finish()
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_async(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test asynchronous call with thinking=True to verify reasoning content."""

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": True})
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call(PROMPT)
        with llm.model(model_id, thinking=llm.ThinkingConfig(level="minimal")):
            response = await response.resume(RESUME_PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_async_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test async streaming call with thinking=True to verify reasoning content."""

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": True})
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call.stream(PROMPT)
        await response.finish()
        with llm.model(model_id, thinking=llm.ThinkingConfig(level="minimal")):
            response = await response.resume(RESUME_PROMPT)
        await response.finish()
        snap.set_response(response)


# Tests for thinking enabled but include_thoughts=False
# Note: The snapshot tests will verify that no thought content appears in the response.
# If thoughts were included, the snapshots would show them and the tests would fail
# on subsequent runs when comparing against the recorded cassettes.


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_thinking_no_include_thoughts_sync(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test synchronous call with thinking enabled but include_thoughts=False.

    This verifies that thoughts are filtered out from the response when
    include_thoughts is False, even though thinking is enabled.
    """

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": False})
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call(PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_thinking_no_include_thoughts_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test streaming call with thinking enabled but include_thoughts=False.

    This verifies that thought chunks are not yielded during streaming when
    include_thoughts is False.
    """

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": False})
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call.stream(PROMPT)
        response.finish()
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_no_include_thoughts_async(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test asynchronous call with thinking enabled but include_thoughts=False.

    This verifies that thoughts are filtered out from the response when
    include_thoughts is False.
    """

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": False})
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call(PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_no_include_thoughts_async_stream(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test async streaming call with thinking enabled but include_thoughts=False.

    This verifies that thought chunks are not yielded during async streaming when
    include_thoughts is False.
    """

    @llm.call(model_id, thinking={"level": "medium", "include_thoughts": False})
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call.stream(PROMPT)
        await response.finish()
        snap.set_response(response)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:completions",
            "provider_model_name": "gpt-4o:completions",
            "params": {"thinking": {"level": "none"}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 91,
                "output_tokens": 23,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "CompletionUsage(completion_tokens=23, prompt_tokens=91, total_tokens=114, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
                "total_tokens": 114,
            },
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            args='{"n": 100}',
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {"n": 100},
                                    "name": "compute_fib",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n\xe2\x01\x01r\xc8\xda|\xa13\x03\xda\xb1whX\xf5(;\xc7,^\xf3n\x9a4\xe5\x9b^\xa3\xa3\xdd\xdf\t\xae\x03fz\xed^l\x06:\x8f\t\x94\x1fIop^\xa0\x1e\x1eO\x17w\xe0EA\xd0\x16\xa1\xc2\xe3\xd35\xcbz\xc6\xf6B\xa2T\xf1h\xde\x88\xe3\xa4%\xa7[\xe9\xeaO\xa7\x0egz\xa2\xd8\xb7\xd6\x86\xbeA\xcaB\xd6b\x9eu@\xf4\xb5\x80\x14\x04)\xb8}\xb2\xce\xc4(\x88\x81\xf1\xa8\xf8\xc5\xc7\x08V\xb1\xe22k\x87\xfb\x91^5S&\xe7\xd8i\xe1\xac\xa0\x80;B\xad\xbc\x96\xa1's\xb9tk\xafz9E\xfatF\x86\x08\x02\xe6J\x13\xdb\xdd\xc2\xd5\xb9\xc2Ji\xce\xf4\x06\xa0\xaffL\xce\xdd 8\x02\xe9Xz\x8eg\xb5\xb6\x9f\xd7\x97\xde\xef\xe6xf!I>\xa1,\xf0y\xe76\x86(\xcc\x86\xf9\x15\x9d\xfca\x9a\xbd\xc0\xb3\x86d+\xff\xd0",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            result="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218,922,995,834,555,169,026."
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:completions",
                    provider_model_name="gpt-4o:completions",
                    raw_message={
                        "content": "The 100th Fibonacci number is 218,922,995,834,555,169,026.",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        }
    }
)



"""
--------------------------------------------------------------------------------
Source: https://github.com/pydantic/pydantic/issues/6381#issuecomment-1831607091
By: silviumarcu

This code is used in accordance with the repository's license, and this reference
serves as an acknowledgment of the original author's contribution to this project.
--------------------------------------------------------------------------------
"""

import inspect
from typing import Any, Generic, NoReturn, Union, cast, get_args, get_origin

from pydantic import BaseModel, create_model

from .format import FormattableT

# Cache for generated partial models to avoid recreation
_partial_model_cache: dict[type[Any], type[Any]] = {}


class Partial(Generic[FormattableT]):
    """Generate a new class with all attributes optionals.

    Notes:
        This will wrap a class inheriting form BaseModel and will recursively
        convert all its attributes and its children's attributes to optionals.

    Example:
        Partial[SomeModel]
    """

    def __new__(
        cls,
        *args: object,  # noqa :ARG003
        **kwargs: object,  # noqa :ARG003
    ) -> "Partial[FormattableT]":
        """Cannot instantiate.

        Raises:
            TypeError: Direct instantiation not allowed.
        """
        raise TypeError(
            "Cannot instantiate abstract Partial class."
        )  # pragma: no cover

    def __init_subclass__(
        cls,
        *args: object,
        **kwargs: object,
    ) -> NoReturn:
        """Cannot subclass.

        Raises:
           TypeError: Subclassing not allowed.
        """
        raise TypeError(f"Cannot subclass {cls.__module__}.Partial")  # pragma: no cover

    def __class_getitem__(
        cls,
        wrapped_class: type[FormattableT],
    ) -> type[FormattableT]:
        """Convert model to a partial model with all fields being optionals.

        Recursively converts all fields in a Pydantic BaseModel to optional,
        handling nested models and generic types like list[Book].

        Args:
            wrapped_class: The BaseModel class to make partial

        Returns:
            A new BaseModel class with all fields optional (or original if not BaseModel)

        Example:
            >>> class Author(BaseModel):
            ...     first_name: str
            ...     last_name: str
            >>> class Book(BaseModel):
            ...     title: str
            ...     author: Author
            >>> PartialBook = Partial[Book]
            >>> partial = PartialBook(title="The Name")
            >>> partial.author  # None
        """
        # Return non-BaseModel types unchanged
        if not (
            inspect.isclass(wrapped_class) and issubclass(wrapped_class, BaseModel)
        ):
            return wrapped_class

        # Check cache to avoid regenerating
        if wrapped_class in _partial_model_cache:
            return cast(type[FormattableT], _partial_model_cache[wrapped_class])

        # Recursively make all fields optional
        partial_fields: dict[str, Any] = {}
        for field_name, field_info in wrapped_class.model_fields.items():
            field_type = field_info.annotation

            # Recursively handle nested BaseModel fields
            if inspect.isclass(field_type) and issubclass(field_type, BaseModel):
                field_type = Partial[field_type]

            # Handle generic types with BaseModel args (e.g., list[Book], dict[str, Book])
            origin = get_origin(field_type)
            if origin is not None:
                args = get_args(field_type)
                # Recursively convert BaseModel args to partial
                new_args = tuple(
                    Partial[arg]
                    if inspect.isclass(arg) and issubclass(arg, BaseModel)
                    else arg
                    for arg in args
                )
                # Reconstruct generic type with new args
                if new_args != args:
                    field_type = origin[new_args]

            # Make field optional with None default
            optional_type = Union[field_type, None]  # noqa: UP007
            partial_fields[field_name] = (optional_type, None)

        # Create new model with "Partial" prefix
        partial_model = create_model(
            f"Partial{wrapped_class.__name__}", __base__=BaseModel, **partial_fields
        )

        # Cache the generated model
        _partial_model_cache[wrapped_class] = partial_model

        return cast(type[FormattableT], partial_model)



import math

from mirascope import llm


@llm.tool
def sqrt_tool(number: float) -> float:
    """Computes the square root of a number"""
    return math.sqrt(number)


@llm.prompt(tools=[sqrt_tool])
def math_assistant(query: str):
    return query


model = llm.Model("openai/gpt-5-mini")
response = math_assistant(model, "What's the square root of 4242?")

while response.tool_calls:
    tool_outputs = response.execute_tools()
    response = response.resume(tool_outputs)

print(response.text())



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "response1": '"General Kenobi!" A surprise to be sure, but a welcome one. How may I assist you on your galactic journey today?',
        "usage1": {
            "input_tokens": 3659,
            "output_tokens": 28,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=28, prompt_tokens=3659, total_tokens=3687, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
            "total_tokens": 3687,
        },
        "response2": "\"Ah, the most important meal of the day, it is! Start your morning like a true Jedi with a nutritious breakfast that even Yoda would approve of. How about some Bantha blue milk pancakes with a side of roasted Porg eggs? They're sure to give you the energy of an entire squad of Clone Troopers! If you're feeling more like a scoundrel today, opt for a stack of Wookiee-approved Waffles with Kowakian monkey-lizard syrup. Punch it, Chewie, and remember: 'This is the Way' to start your day!\"",
        "usage2": {
            "input_tokens": 3663,
            "output_tokens": 117,
            "cache_read_tokens": 3584,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=117, prompt_tokens=3663, total_tokens=3780, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=3584))",
            "total_tokens": 3780,
        },
    }
)
